<img src="ml.jpg" width=2000/><br>
# Exercise Sheet 2 - Machine Learning
Author: Samuel Innes <br>
Email: dd257@stud.uni-heidelberg.de

## <a href="_02_EXPML_ML_supervised.pdf">Link to Exercise sheet</a>

## :scroll: Step-by-Step Description
1. <em>(Lines 9-22)</em>: Functions imported from [numpy](https://numpy.org/), [matplotlib](https://matplotlib.org/) and [scikit-learn](https://scikit-learn.org/stable/).
2. <em>(Lines 26-37)</em>: Classifiers are initialized, and a list of their names created (for the graphical representation).
3. <em>(Lines 39-48)</em>: 3 random datasets are created with differen patterns: 2 moons, 2 circles and 2 strips (linear).
4. <em>(Lines 53-81)</em>: Iterate over list of datasets, 
    1. <em>(Lines 54-63)</em>: define the training and test sets, 
    2. <em>(Lines 66-70)</em>: plot input training (more opaque) and test (more transparent) data points on a graph,
    3. <em>(Lines 84-115)</em>: iterate over classifiers,
        1. <em>(Lines 85-86)</em>: train the algorithms on the training set,
        2. <em>(Line 87)</em>: evaluate classifier on test set and calculate score,
        3. <em>(Lines 97-115)</em>: plot data points and classifier regions on the graphs, with title and accuracy score.
5. <em>(Lines 117-118)</em>: Show graphs in a readable (tight_layout) way.

## :clock130: Execution time
Using 
```python
import time
start = time.time()
#[CODE]
end = time.time()
print(end-start)
```
to calculate the time taken each iteration, then adding them to a list of times. To update the fastest and slowest times, this code was executed at the end of each iteration of the loop:
```python
times_list.append(exc_time)
if exc_time<fastest_time:
    fastest_time = exc_time
if exc_time>slowest_time:
    slowest_time = exc_time
```
Then to calculate the metrics, I used the following code outside the loop:
```python
average_time = sum(times_list)/len(times_list)
print("Fastest time: " + str(fastest_time))
print("Slowest time: " + str(slowest_time))
print("Average time: " + str(average_time))
print("Standard deviation: " + str(numpy.std(times_list, dtype=np.float64)))
``` 
## Results
Execution time range (based on **10** runs): <br>
Fastest time: **5.979996919631958**<br>
Slowest time: **6.943421125411987**<br>
Average time: **6.339807081222534**<br>
Standard deviation: **0.34304171431823993**<br><br>
The execution time is slightly different each run, as each time the processor and RAM are in a different state, which has a small but noticeable impact on the speed of the program. <br><br>
Therefore, when we are developing or testing machine learning algorithms, if we are interested in the speed of the algorithm, we should repeat the execution several times to calculate an average. In order to compare different algorithms we also must make sure they have been tested on the same computer with the same hardware in order to get a reliable result.

##### :computer: Hardware
- Processor: Intel(R) Core(TM) i3-6006U CPU @ 2.00GHz   1.99 GHz
- Graphics card: Intel(R) HD Graphics 520
- RAM: 8.00 GB (7.89 GB usable)
- SSD: SanDisk SD8SN8U-256G

## :chart_with_upwards_trend: Algorithms and data

#### 1. What characterizes the different datasets?
In the first dataset, the data points are in a moon formation, so that they intertwine in the middle. 
<br>![](moon-data.JPG)<br>
In the second, they are arranged in two rings of different sizes, but with the same center. 
<br>![](ring-data.JPG)<br>
And in the third, they are in two stripes.
<br>![](strip-data.JPG)<br>

#### 2. Nature of decision boundaries
- Nearest Neighbours has a relatively sharply graded decision boundary, with two "levels": a core domain and a borderline area for each classification. The boundary lines themselves are non-linear.
- Linear SVM has a much more softly graded boundary, with several "levels of confidence" for each classifier, where the most confident classification is far away from the centers of other classes. The  boundary lines are straight.
- Kernel SVM is similar to Linear SVM in the graded boundaries, however the boundary lines are curved and based around a kernel center, which is an area with the highest density of data points of a particular class. 
- Decision Tree has extremely starkly defined areas with little to no "uncertainty" or gradation of classification. The boundary lines are also linear and straight. However, unlike Linear SVM, a domain can be defined by several lines perpendicular to each other. 
- Random Forest has similar boundary lines to Decision Tree, however with a graded level of certainty of classification. 
- Neural Net is similar to Kernel SVM in respect to the amount of gradation, however instead of domains being focussed around a kernel center with boundary lines delineating the area, the domain is defined by a combination of boundary lines and extends beyond the region of the data sets, to the axis-boundaries and beyond. 
- Naive Bayes has a similar result to Neural Net in the shape and gradation of the boundaries, however the form of the boundary lines tends more towards a polynomial structure. 

#### 3. Linear Separability
"Linearly separable" means there exists a line on the plain, where all the points in one class are on the one side, and all points in the other class on the other. <br>
Of the given datasets, technically none are linearly separable, however, taking into accound a certain amount leeway into the results, it can certainly be argued that the last dataset is linearly separable.
