import numpy as np
import matplotlib.pyplot as plt
from matplotlib import rc 
from matplotlib.colors import ListedColormap
from sklearn.datasets import make_moons, make_circles, make_classification
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import StandardScaler
from sklearn.tree import DecisionTreeClassifier, plot_tree
from sklearn.ensemble import RandomForestClassifier
font = {'family' : 'normal',
        'weight' : 'bold',
        'size'   : 4}

rc('font', **font)

h = 0.02 # step size in the mesh
np.random.seed(1)
# create datasets
X, y = make_classification(
    n_features=2, n_redundant=0, n_informative=2, random_state=1, n_clusters_per_class=1
)
rng = np.random.RandomState(2)
X += 2 * rng.uniform(size=X.shape)
linearly_separable = (X, y)

datasets = {
    "moons": make_moons(noise=0.3, random_state=0),
    "circles": make_circles(noise=0.2, factor=0.5, random_state=1),
    "linearly separable": linearly_separable,
}



for k in [0, 1]:
    #for each dataset, train, predict test data and calculate accuracy
    i = 0
    fig, ax = plt.subplots(3, 5)
    for ds_name, ds in datasets.items():
        # split into test and train sets
        X, y = ds
        X = StandardScaler().fit_transform(X)
        X_train, X_test, y_train, y_test = train_test_split(
            X, y, test_size=0.4, random_state=42
        )
        x_min, x_max = X[:, 0].min() - 0.5, X[:, 0].max() + 0.5
        y_min, y_max = X[:, 1].min() - 0.5, X[:, 1].max() + 0.5
        xx, yy = np.meshgrid(np.arange(x_min, x_max, h), np.arange(y_min, y_max, h))

        cm = plt.cm.RdBu
        cm_bright = ListedColormap(["#FF0000", "#0000FF"])
        ax[i,0].scatter(X_train[:, 0], X_train[:, 1], c=y_train, cmap=cm_bright, edgecolors="k")
        ax[i,0].scatter(X_test[:, 0], X_test[:, 1], c=y_test, cmap=cm_bright, edgecolors="k")
        ax[i,0].set_xlim(xx.min(), xx.max())
        ax[i,0].set_ylim(yy.min(), yy.max())
        ax[i,0].set_xticks(())
        ax[i,0].set_yticks(())

        #set titles
        ax[0,1].set_title('default max_depth X_train')
        ax[0,2].set_title('default max_depth X_test')
        ax[0,3].set_title("max_depth = optimum X_train")
        ax[0,4].set_title("max_depth = optimum X_test")
        
        #with default max depth
        if k == 0:
            cls = DecisionTreeClassifier().fit(X_train, y_train)
        elif k == 1:
            cls = RandomForestClassifier().fit(X_train, y_train)

        print("\nDataset type: " + ds_name)
        print("Prediction on test set" + str(cls.predict(X_test)))
        print("Accuracy on test data:" + str(cls.score(X_test, y_test)))
        print("Accuracy on training data:" + str(cls.score(X_train, y_train)))

        ax[i,1].scatter(X_train[:, 0], X_train[:, 1], c=y_train, cmap=cm_bright, edgecolors="k")
        ax[i,2].scatter(X_test[:, 0], X_test[:, 1], c=y_test, cmap=cm_bright, edgecolors="k", alpha=0.5)
        ax[i,1].set_xlim(xx.min(), xx.max())
        ax[i,2].set_xlim(xx.min(), xx.max())
        ax[i,1].set_ylim(yy.min(), yy.max())
        ax[i,2].set_ylim(yy.min(), yy.max())
        ax[i,1].set_xticks(())
        ax[i,2].set_xticks(())
        ax[i,1].set_yticks(())
        ax[i,2].set_yticks(())

        #plot decision boundary
        Z = cls.predict_proba(np.c_[xx.ravel(), yy.ravel()])[:,1]
        Z = Z.reshape(xx.shape)
        ax[i,2].contourf(xx, yy, Z, cmap=cm, alpha=0.8)

        #with optimum max depth
        if k == 0:
            cls = DecisionTreeClassifier(max_depth=5).fit(X_train, y_train)
        elif k == 1:
            cls = RandomForestClassifier(max_depth=7).fit(X_train, y_train)

        print("\nDataset type: " + ds_name)
        print("Prediction on test set" + str(cls.predict(X_test)))
        print("Accuracy on test data:" + str(cls.score(X_test, y_test)))
        print("Accuracy on training data:" + str(cls.score(X_train, y_train)))

        ax[i,3].scatter(X_train[:, 0], X_train[:, 1], c=y_train, cmap=cm_bright, edgecolors="k")
        ax[i,4].scatter(X_test[:, 0], X_test[:, 1], c=y_test, cmap=cm_bright, edgecolors="k", alpha=0.5)
        ax[i,3].set_xlim(xx.min(), xx.max())
        ax[i,4].set_xlim(xx.min(), xx.max())
        ax[i,3].set_ylim(yy.min(), yy.max())
        ax[i,4].set_ylim(yy.min(), yy.max())
        ax[i,3].set_xticks(())
        ax[i,4].set_xticks(())
        ax[i,3].set_yticks(())
        ax[i,4].set_yticks(())

        #plot decision boundary
        Z = cls.predict_proba(np.c_[xx.ravel(), yy.ravel()])[:,1]
        Z = Z.reshape(xx.shape)
        ax[i,4].contourf(xx, yy, Z, cmap=cm, alpha=0.8)

        if k == 0:
            figt, axt = plt.subplots(1,1)
            plot_tree(cls, filled=True)

        i += 1

plt.tight_layout()
plt.show()