<img src="images/birch-forest.jpg" width=2000/>[^1]

# Slavic language classifier
This project takes as input a text in a particular Slavic language and returns which language it is, including information about the linguistic classification of that language. 

### Contents:
- <a href="https://gitlab.cl.uni-heidelberg.de/innes/exp-ml-1/-/blob/master/project/README.md#background">Background</a>
    - <a href="https://gitlab.cl.uni-heidelberg.de/innes/exp-ml-1/-/blob/master/project/README.md#motivation">Motivation</a>
    - <a href="https://gitlab.cl.uni-heidelberg.de/innes/exp-ml-1/-/blob/master/project/README.md#similar-projects">Similar Projects</a>
    - <a href="https://gitlab.cl.uni-heidelberg.de/innes/exp-ml-1/-/blob/master/project/README.md#the-slavic-language-family">The Slavic language family</a>
- <a href="https://gitlab.cl.uni-heidelberg.de/innes/exp-ml-1/-/blob/master/project/README.md#model">Model</a>
    - <a href="https://gitlab.cl.uni-heidelberg.de/innes/exp-ml-1/-/blob/master/project/README.md#supported-languages">Supported languages</a>
    - <a href="https://gitlab.cl.uni-heidelberg.de/innes/exp-ml-1/-/blob/master/project/README.md#corpus">Corpus</a>
    - <a href="https://gitlab.cl.uni-heidelberg.de/innes/exp-ml-1/-/blob/master/project/README.md#features">Features</a> 
    - <a href="https://gitlab.cl.uni-heidelberg.de/innes/exp-ml-1/-/blob/master/project/README.md#label-frequencies">Label frequencies</a> 
    - <a href="https://gitlab.cl.uni-heidelberg.de/innes/exp-ml-1/-/blob/master/project/README.md#algorithm">Algorithm</a>
- <a href="https://gitlab.cl.uni-heidelberg.de/innes/exp-ml-1/-/blob/master/project/README.md#development">Development</a>
    - <a href="https://gitlab.cl.uni-heidelberg.de/innes/exp-ml-1/-/blob/master/project/README.md#parameters">Parameters</a>
      - <a href="https://gitlab.cl.uni-heidelberg.de/innes/exp-ml-1/-/blob/master/project/README.md#decision-tree-and-random-forest">Decision Tree and Random Forest</a>
      - <a href="https://gitlab.cl.uni-heidelberg.de/innes/exp-ml-1/-/blob/master/project/README.md#neural-network">Neural Network</a>
      - <a href="https://gitlab.cl.uni-heidelberg.de/innes/exp-ml-1/-/blob/master/project/README.md#naive-bayes">Naive Bayes</a>
    - <a href="https://gitlab.cl.uni-heidelberg.de/innes/exp-ml-1/-/blob/master/project/README.md#feature-importance">Feature importance</a>
- <a href="https://gitlab.cl.uni-heidelberg.de/innes/exp-ml-1/-/blob/master/project/README.md#performance">Performance</a>
    - <a href="https://gitlab.cl.uni-heidelberg.de/innes/exp-ml-1/-/blob/master/project/README.md#baselines">Baselines</a>
    - <a href="https://gitlab.cl.uni-heidelberg.de/innes/exp-ml-1/-/blob/master/project/README.md#evaluation">Evaluation</a>
- <a href="https://gitlab.cl.uni-heidelberg.de/innes/exp-ml-1/-/blob/master/project/README.md#technical-details">Technical details</a>
    - <a href = "https://gitlab.cl.uni-heidelberg.de/innes/exp-ml-1/-/blob/master/project/README.md#system-requirements">System Requirements</a>
    - <a href = "https://gitlab.cl.uni-heidelberg.de/innes/exp-ml-1/-/blob/master/project/README.md#installation">Installation</a>
    - <a href = "https://gitlab.cl.uni-heidelberg.de/innes/exp-ml-1/-/blob/master/project/README.md#author">Author</a>
- <a href="https://gitlab.cl.uni-heidelberg.de/innes/exp-ml-1/-/blob/master/project/README.md#references">References</a>

# Background

## Motivation
Although many language detectors already exist, there are none which focus specifically on the Slavic language family and are thus able to detect less widely-spoken Slavic languages such as Lower Sorbian or Rusyn. However, using the fact that Wikipedia is available in many of this smaller languages, a classifier can be trained, as is explained below, to detect these minority languages. Such a classifier can help towards preservation of these languages and can contribute to their digital presence, thus encouraging more research into them and into the development of NLP tools for their speakers. This in turn makes it easier to use these languages and thus helps them to survive in a digital age.

## Similar projects
The most obvious and popular example for language recognition is <a href="https://translate.google.co.uk/">Google Translate</a>. The amount of languages offered is ever growing, however of the 16 supported languages listed below, Google Translate can only detect 10. <a href="https://www.deepl.com/translator">DeepL</a> is another Machine Translation tool that has become popular in recent years, however has an even smaller selection of languages. 

There are other specialist websites offering language classification, such as <a href="https://www.dcode.fr/language-recognition">dcode.fr</a> or <a href="whatlanguageisthis.com">whatlanguageisthis.com</a>, however these also have no support for lesser spoken Slavic languages, with an even smaller range than Google Translate.

## The Slavic language family
The Encyclopedia Britannica describes the Slavic language family thus:
><i>The Slavic languages, also called Slavonic languages, are a group of Indo-European languages spoken in most of Eastern Europe, much of the Balkans, parts of Central Europe, and the northern part of Asia. The Slavic languages, spoken by some 315 million people at the turn of the 21st century, are most closely related to the languages of the Baltic group (Lithuanian, Latvian, and the now-extinct Old Prussian), but they share certain linguistic innovations with the other eastern Indo-European language groups (such as Indo-Iranian and Armenian) as well.<i>[^2]

<img src="images/Slavic_europe.jpg" width=400>

[^3]

I chose this area mainly because of personal interest: the Slavic or, more generally, Eastern European world is one that greatly interests me and is hence my second subject at university. It is therefore a personal goal of mine to help preserve the linguistic diversity of this region. This project could however very easily be implemented on other language groups with an available corpus or with a Wikipedia.

<img src="images/birch-forest-crop.jpg" width=2000/>

# Model

## Supported languages
The list of Wikipedia sites in Slavic languages is as follows:
Russian, Polish, Ukrainian, Slovenian, Slovakian, Bulgarian, Belarussian, Czech, Croatian, Macedonian, Serbian*, Serbo-Croatian*, Bosnian*, Silesian, Upper Sorbian, Lower Sorbian, Rusyn, Old Church Slavonic (in Cyrillic and Glagolitic script)*.

However, due to technical reasons, it was necessary to make the following adjustments to the list:
<ul>
  <li>Due to the linguistic similarity of the Serbo-Croatian language group and in order to avoid inadvertent political statements, the individual languages/dialects of Serbian, Croatian and Bosnian have been replaces with two categories: Serbo-Croatian: Cyrillic and Serbo-Croatian: Latin.
  <li>Old Church Slavonic Wikipedia is also available in two alphabets: Cyrillic and Glagolitic. However, the Wikipedia API only supports Cyrillic, hence only this alphabet has been included.
</ul>

Therefore the labels included in the model are the following (in rough order of Wikipedia size):
| Supported languages: | 
| --- | 
Russian
Polish
Ukrainian
Slovenian
Slovakian
Bulgarian
Belarussian
Czech
Macedonian
Serbo-Croatian: Cyrillic
Serbo-Croatian: Latin
Silesian
Upper Sorbian
Lower Sorbian
Rusyn
Old Church Slavonic: Cyrillic

Here is a visualisation of the Slavic language family, a subfamily of Balto-Slavic languages, with its different branches.

<img src="images/lang_tree.svg" width=700><br>[^4]

## Corpus
Using webscraping to extract text from Wikipedia, it is possible to create a reliable text corpus, even for lesser spoken Slavic languages. For a more in-depth discussion of the use of Wikipedia as a corpus, see <a href="PasternackRo08.pdf">Pasternak & Roth 2008</a>. In order to scrape the corpus from Wikipedia, the `wikipedia` <a href="https://wikipedia.readthedocs.io/en/latest/">python library</a> was used. This is much faster and saves a lot of the HTML-parsing work, which would have otherwise been necessary with classic web-scraping. The library also contains useful functions such as `wikipedia.set_lang()` to set the language and `wikipedia.random(n)` to return a list of n random article titles, as well as the class `wikipedia.WikipediaPage` with the attribute `.content` to return the content of an article as a string.

Thus by randomly finding 500 articles (the highest number `wikipedia.random()` allows), which, in order to avoid overfitting, are immediately split into 2 parts: one for Train/Dev and the other for the Test set, one can create a corpus (Train/Dev) for feature extraction. Cross-Validation was used since the corpus is relatively small.

## Features
Originally the project began by using a Bag of Words model (BoW). The reasons why a BoW model is the most suitable for this task are as follows: three of the most informative features for language classification are the following: vocabulary, orthography and morphology. This is especially the case in the Slavic language group due to its high level of synthesis[^5] and thus of syntactic flexibility. All three features are present in a Bag of Words model, since it is simply written language (orthography), tokenised (vocabulary/lexis and morphology) and then each token is counted.

<img src="images/bow.png" width=500><br>[^6]

The smallest Wikipedia in a Slavic language is Old Church Slavonic with 846 articles, a comparatively very small corpus, and other languages have a similarly small number of articles. This had an impact on accuracy during training and development: using `cross_val_score()` to calculate performance on the train/dev data, the model had an accuracy of `0.95`. However, when predicting on texts from a source other than Wikipedia, the accuracy dropped to around the `0.50` mark. This is because the model had overfitted to vocabulary common on Wikipedia pages such as _born, year_ or _from_, which are part of a template used to create them.

A common alternative to BoW is the more sophisticated n-gram model, where sequences of tokens (of length n) are counted. However, this method naturally requires a much larger corpus for training and carries a lot of syntactic information, which is, as shall be elaborated later, of less importance than other linguistic features.

A rethink was therefore necessary, and it was decided to return to what features _humans_ use mostly when trying to detect an unknown language. Inspired by <a href="https://op.europa.eu/en/publication-detail/-/publication/1c672bb8-ffa8-11e5-b713-01aa75ed71a1">this publication by the EU</a>, a guide to language identification, the most obvious feature is, of course, the alphabet. The chosen Slavic languages also lend themselves to this, since the overwhelming majority have characters unique to the language. Although this is already to a certain extent encoded into the BoW model, with such a small corpus, its significance is rather low, and it makes more sense to input the letters themselves as features. <a href = "alphabets">All the individual alphabets were therefore written out by hand as separate sets</a>, merging them to avoid redundant features (see code <a href = "alphabets">here</a>).

The features are therefore the n most common words in the language (many of which will be stop words) combined with the frequency of the letters in <a href = "combined_alphabet.txt">`combined_alphabet.txt`</a>.

Some example document vectors could look something like this:

<table>
<tr>
  <td colspan="4"><b>Russian</b></td>
  <td colspan="4"><b>Bulgarian</b></td>
  <td>...</td>
  <td colspan="4"><b>Polish</b></td>
  <td colspan="4"><b>Letters</b></td>
</tr>
<tr>
  <!--Russian words-->
  <td>и</td>
  <td>в</td>
  <td>не</td>
  <td>...</td>
  <!--Bulgarian words-->
  <td>да</td>
  <td>се</td>
  <td>на</td>
  <td>...</td>
  <!--...-->
  <td>...</td>
  <!--Polish words-->
  <td>nie</td>
  <td>to</td>
  <td>się</td>
  <td>...</td>
  <!--Letters-->
  <td>а</td>
  <td>б</td>
  <td>...</td>
  <td>đ</td>
</tr>
<tr>
  <!--Russian words-->
  <td>0</td>
  <td>0</td>
  <td>1</td>
  <td>...</td>
  <!--Bulgarian words-->
  <td>5</td>
  <td>3</td>
  <td>5</td>
  <td>...</td>
  <!--...-->
  <td>...</td>
  <!--Polish words-->
  <td>0</td>
  <td>0</td>
  <td>0</td>
  <td>...</td>
  <!--Letters-->
  <td>45</td>
  <td>23</td>
  <td>...</td>
  <td>0</td>
</tr>
<tr>
  <!--Russian words-->
  <td>2</td>
  <td>0</td>
  <td>0</td>
  <td>...</td>
  <!--Bulgarian words-->
  <td>1</td>
  <td>0</td>
  <td>0</td>
  <td>...</td>
  <!--...-->
  <td>...</td>
  <!--Polish words-->
  <td>0</td>
  <td>0</td>
  <td>0</td>
  <td>...</td>
  <!--Letters-->
  <td>1</td>
  <td>0</td>
  <td>...</td>
  <td>0</td>
</tr>
<tr>
  <!--Russian words-->
  <td>7</td>
  <td>9</td>
  <td>8</td>
  <td>...</td>
  <!--Bulgarian words-->
  <td>1</td>
  <td>2</td>
  <td>5</td>
  <td>...</td>
  <!--...-->
  <td>...</td>
  <!--Polish words-->
  <td>0</td>
  <td>0</td>
  <td>0</td>
  <td>...</td>
  <!--Letters-->
  <td>33</td>
  <td>17</td>
  <td>...</td>
  <td>1</td>
</tr>
<tr>
  <!--Russian words-->
  <td>0</td>
  <td>0</td>
  <td>0</td>
  <td>...</td>
  <!--Bulgarian words-->
  <td>0</td>
  <td>0</td>
  <td>0</td>
  <td>...</td>
  <!--...-->
  <td>...</td>
  <!--Polish words-->
  <td>5</td>
  <td>6</td>
  <td>3</td>
  <td>...</td>
  <!--Letters-->
  <td>0</td>
  <td>0</td>
  <td>...</td>
  <td>0</td>
</tr>
<tr>
  <!--Russian words-->
  <td>1</td>
  <td>0</td>
  <td>1</td>
  <td>...</td>
  <!--Bulgarian words-->
  <td>0</td>
  <td>1</td>
  <td>0</td>
  <td>...</td>
  <!--...-->
  <td>...</td>
  <!--Polish words-->
  <td>0</td>
  <td>1</td>
  <td>0</td>
  <td>...</td>
  <!--Letters-->
  <td>5</td>
  <td>0</td>
  <td>...</td>
  <td>45</td>
</tr>
<tr>
  <!--Russian words-->
  <td>0</td>
  <td>0</td>
  <td>0</td>
  <td>...</td>
  <!--Bulgarian words-->
  <td>0</td>
  <td>0</td>
  <td>0</td>
  <td>...</td>
  <!--...-->
  <td>...</td>
  <!--Polish words-->
  <td>0</td>
  <td>1</td>
  <td>0</td>
  <td>...</td>
  <!--Letters-->
  <td>2</td>
  <td>0</td>
  <td>...</td>
  <td>0</td>
</tr>
</table>

## Label frequencies
Since the corpus was specially created for this project, the label frequencies could be easily controlled, and therefore, in order to avoid the problems that come with an imbalanced dataset, the dataset was created as a balanced one. The maximum number of randomly generated articles with the `wikipedia.random(n)` method from the Wikipedia API is 500 and thus this number was chosen as the number of data points per language. Since some of the randomly fetched articles were either empty, just an image or had other issues meaning the content text couldn't be extracted, the label frequencies have a slight variance. Here is a histogram of the label frequencies (see code <a href="development/calculate_label_frequencies.py">here</a>):

<img src="images/label_hist.jpeg" width=400>

And here are the absolute values:
| Language | No. data points |
| --- | --- |
Rusyn | 500 
Silesian | 497
Macedonian | 496
Lower Sorbian | 493
Old Church Slavonic | 490
Slovene | 486
Upper Sorbian | 483
Serbo-Croatian: Latin | 482
Bulgarian | 481
Ukrainian | 479
Belarussian | 476
Polish | 468
Czech | 467
Serbo-Croatian: Cyrillic | 459
Russian | 447
Slovakian | 434

Thus, as is clear from both the histogram and the table, the dataset is relatively (though not completely) balanced - the standard deviation is `17.91` - however each data point is only annotated with one label.

## Algorithm
Firstly it was decided that Decision Tree/Random Forest would be a good approach. Since the languages themselves are grouped in a tree-structure and with a locally greedy approach, a Decision Tree/Random Forest could in theory model the historical linguistic development and development of different features well, thus modelling the structure of the problem itself. The idea that the Decision Tree could model the linguistic family tree entails the assumption that similar languages have similar words and similar alphabets, which is true in many cases however not in all. 

In practice, this proved impractical due to the large number of classes and features. This cutout gives an idea of the complexity and thus impracticality of the resulting Decision Tree:

<img src="images/decision_tree.png" width=800/>

However, both the Decision Tree and the Random Forest classifiers achieved a relatively high accuracy (see <a href="https://gitlab.cl.uni-heidelberg.de/innes/exp-ml-1/-/blob/master/project/README.md#evaluation">Evaluation</a>), with the latter predictably being the best of the two on the test set. Since the problem is in effect relatively simple and the number of classes high, a Random Forest classifier is a very good choice.

With its flexibility and capability of highlighting certain feature combinations, a Multi-Layer Perceptron was also trained on the data. And as is explained later, after optimising the parameters (see <a href="https://gitlab.cl.uni-heidelberg.de/innes/exp-ml-1/-/blob/master/project/README.md#development">Development</a>), it lead to the best results on the Test set, despite how small the dataset is. However the training-time was noticeably longer. 

For comparison, a Naive Bayes model was also trained on the data to use as a baseline. For a discussion of the results of the different algorithms see <a href="https://gitlab.cl.uni-heidelberg.de/innes/exp-ml-1/-/blob/master/project/README.md#evaluation">Evaluation</a>.

<img src="images/birch-forest-crop.jpg" width=2000/>

# Development

## Parameters

### Decision Tree and Random Forest
See <a href = "development/DTRF_optimise_param.py">DTRF_optimise_param.py</a> for the source code in full.<br>

After first shuffling the data using `shuffle(X, y, random_state=0)`, it was split into train/test sets using:
```python
from sklearn.model_selection import train_test_split
X_train_and_dev, X_test, y_train_and_dev, y_test = \
  train_test_split(shuffled_X, shuffled_y, test_size=0.25, random_state=0) # first separate test data
```

This was saved in `train_data.txt` and `test_data.txt` for more clarity.

Due to the small corpus, sklearn's `cross_val_score(rf, X, y, cv=5)` was used, which automatically splits the training data into train and development sets for cross-validation. In order to achieve the 60-15-25 proportions required in the exercise sheet, the number of folds was set as `cv = 5`, since $`\frac{15}{60+15}=0.2`$, meaning each validation fold would have the required proportions. A dev set was also created with `X_train, X_dev, y_train, y_dev = train_test_split(X_train_and_dev, y_train_and_dev, test_size=0.2)`, but only for demonstration purposes. In this case, since the corpus contains 7638 data points, the absolute proportions are: 4583-1146-1909.

Using cross-validation as described above, the effect of different `max_depth` parameters on the accuracy of a Random Forest Classifier was investigated. This was done by taking an average of each fold and plotting this on a graph against the max_depth.

<img src="images/max_depth_accuracy.jpeg" width=400>

In this example, the default (baseline) `max_depth` value was the highest, with an accuracy of `0.9825`, thus the default value was used. 

A hyperparameter unique to the Decision Tree is `splitting`, which essentially decides which features to base the decision on (random or best). Using the random setting can help against overfitting, however in this case, it had a lower average accuracy of `0.9203901933438503` compared with the default (`"best"`) of `0.9418646974096344`.

`min_samples_split` is another hyperparameter which can be tuned. Using a for-loop to investigate the impact of changing it on the accuracy, the following graph was created. As is clear, the default performed better than other values for the parameter.

<img src="images/min_samples_split.png" width=400>

Changing other hyperparameters such as `max_features`, `min_samples_split` or `min_samples_leaf` (along with many others) is mainly to avoid overfitting or decrease the computational complexity of the model. Since time is not really a limiting factor, and since the accuracy with the default settings is already very high, it was decided to stick with the default.

### Neural Network
See <a href = "development/MLP_optimise_param.py">MLP_optimise_param.py</a> for the source code in full.<br>

The Multi-Layer Perceptron was optimised using the class `sklearn.model_selection.GridSearchCV`, due to the large number of parameters. The parameters and the values to be tried out were as follows:

```python
parameter_space = {
    'hidden_layer_sizes': [(50,100,50), (100,)],
    'activation': ['tanh', 'relu'],
    'solver': ['lbfgs', 'adam'],
    'alpha': [0.0001, 0.05],
}
```

The optimal parameters found were: `{'activation': 'tanh', 'alpha': 0.0001, 'hidden_layer_sizes': (100,), 'solver': 'adam'}` with a Cross-Validation accuracy of `0.978`. This set-up is identical to the default except for the activation function (by default ReLU).

### Naive Bayes
See <a href = "development/NB_optimise_param.py">NB_optimise_param.py</a> for the source code in full.<br>

The parameters for the Naive Bayes Classifier were optimised in a similar way to the Random Forest and Decision Tree classifiers. In sklearn's MultinomialNB there are only two hyperparameters that can be optimised: `alpha` and `fit/class_prior`. Firstly, the effect of changing `alpha` (the smoothing term) on the accuracy was investigated. The default is 1, but it is clear that there is a negative correlation between the magnitude of the smoothing term and the accuracy. The highest accuracy was achieved with zero-smoothing.

<img src="images/alpha_nb.png" width=400>

This was calculated using an average over 5-fold Cross-Validation, so as to avoid random variations in the data which could have an effect on the performance. However, although `0` gave the highest accuracy, the value was not set to `0`, since if a model with so many features such as this one is better without smoothing, then there is a large danger of overfitting.

<img src="images/alpha_0_1_nb.png" width=400>

Based off of this graph, it was decided that `0.7` would be the optimum parameter to avoid overfitting while also balancing the increase in accuracy by leaving it out.

With regard to the `fit_prior` hyperparameter, since the dataset is already relatively balanced, changing the default to `True` (meaning the classifier assumes a uniform distribution) had little effect on the accuracy. 
```
fit_prior=False: 0.9209125341990747
fit_prior=True: 0.9207380141292667
```

## Feature importance
During the early stages of the project, when character features were added to the model, it was unclear whether the BoW features became redundant.
<table>
<tr>
  <th colspan="6">Average from Cross-Validation (cv=5)</th>
</tr>
<tr>
  <th>Classifier</th>
  <th>Only BoW</th>
  <th>Only Characters</th>
  <th>Combined</th>
</tr>
<tr>
  <td><b>Random Forest:</b></td>
  <td>0.9338</td>
  <td>0.9674</td>
  <td><b>0.9759</b></td>
</tr>
<tr>
  <td><b>Decision Tree:</b></td>
  <td>0.8703</td>
  <td>0.9314</td>
  <td><b>0.9419</b></td>
</tr>
<tr>
  <td><b>Multi-Layer Perceptron:</b></td>
  <td>0.9354</td>
  <td>0.9705</td>
  <td><b>0.9771</b></td>
</tr>
<tr>
  <td><b>Naive Bayes (baseline):</b></td>
  <td>0.7962</td>
  <td><b>0.9228</b></td>
  <td>0.9207</td>
</tr>
</table>

With this table it becomes clear that the characters (alphabets) are more informative than the 10 most frequent words, however the Decision Tree, Random Forest and MLP algorithms are most accurate when these two factors are combined. It is interesting to note, however, that the accuracy of Naive Bayes decreases with the introduction of the BoW features. This could be due to the fact that Naive Bayes assumes that all features are independent of each other (<a href="https://nlp.stanford.edu/IR-book/html/htmledition/properties-of-naive-bayes-1.html">the Naive Bayes assumption</a>). In this case, this is of course not true, and thus, as we can see, the most informative feature combination is when information about the combination of words and characters is used, which a Naive Bayes classifier cannot do. Introducing the BoW features therefore creates more noise, and thus the accuracy decreases.

After the above discussion on which features to include in the model, it is of great interest to see exactly which features are the most "important" for the model. <a href="https://scikit-learn.org/stable/auto_examples/ensemble/plot_forest_importances.html">Using the sklearn built-in attribute `feature_importances_`</a> and the code given <a href="development/feature_importance.py">here</a>, this graph was plotted to show the MDI (Median Decrease of Impurity) of each feature for a Random Forest classifier. Roughly the first half of the features represents the BoW features, and the second half the alphabets.

<img src="images/feature_importances_NOLABELS.png" width=500>

With this graph it is clearly not the case that with the letters the BoW features are obsolete. This would make sense, since not every character in a language's alphabet is seen in each article due to varying lengths, and thus the frequency of certain words naturally contains a lot of information. It corroborates however the results in the table above, that the word-features are less important than the character-features.

<img src="images/birch-forest-crop.jpg" width=2000/>

# Performance

## Baselines
Two baselines were written from scratch: a **Random baseline** and a **Majority baseline** (which can be found <a href ="evaluation/dummy_classifiers.py">here</a>). The Random baseline randomly picks a label based off of the class-probabilities in the training data. The Majority baseline however simply predicts the most common class each time. In the evaluation, the Naive Bayes algorithm was also used as a baseline.

If the model has a lower accuracy than the Random baseline, then it predicts worse than randomly assigning a class. This could mean that the model has learned the wrong features (for example noise). In a binary classification task, the performance could be improved by predicting the opposite class than the algorithm does, however in this case with 16 classes it is impossible. In other cases, the dataset itself must be cleaned to make sure the algorithm learns the right features. 

A Majority baseline, as already explained, always takes the most frequent class. In cases where the dataset is very imbalanced (for example 95-5), this can produce a deceptively high accuracy (95%), even though the classifier hasn't "learnt" anything. One reason for this could be that the data is not separable. This could also be due to the fact that there is an issue with the data itself.

## Evaluation
After generating the dataset, Cross-Validation with 5 folds was carried out immediately to check that the process had gone smoothly. The Multi-Layer Perceptron, Random Forest, Decision Tree and Naive Bayes classifiers gave the following accuracies:

<table>
<tr>
  <th colspan="6">Cross-Validation scores (cv=5)</th>
  <th>Average</th>
</tr>
<tr>
  <td><b>Random Forest:</b></td>
  <td>0.9808</td>
  <td>0.9764</td>
  <td>0.9738</td>
  <td>0.9773</td>
  <td>0.9712</td>
  <td><i>0.9759</i></td>
</tr>
<tr>
  <td><b>Decision Tree:</b></td>
  <td>0.9459</td>
  <td>0.9433</td>
  <td>0.9415</td>
  <td>0.9450</td>
  <td>0.9406</td>
  <td><i>0.9433</i></td>
</tr>
<tr>
  <td><b>Multi-Layer Perceptron:</b></td>
  <td>0.9817</td>
  <td>0.9738</td>
  <td>0.9764</td>
  <td>0.9721</td>
  <td>0.9851</td>
  <td><i><b>0.9778</b></i></td>
</tr>
<tr>
  <td><b>Naive Bayes (baseline):</b></td>
  <td>0.9215</td>
  <td>0.9241</td>
  <td>0.9293</td>
  <td>0.9109</td>
  <td>0.9179</td>
  <td><i>0.9207</i></td>
</tr>
</table>

As expected, MLP, Random Forest and Decision Tree all performed better than the Naive Bayes baseline, with an accuracy well over the initial goal of 90%. What is also clear, is that MLP and Random Forest have a better accuracy than Decision Tree, which was also expected. Also not unexpected was the result of the Neural Network as they are renowned for their ability to achieve high accuracy, combining features in a more complex way than 'simpler' algorithms such as Decision Trees. Although the corpus was small, the MLP performed better than the other classifiers used, and thus the <a href="final_model.zip">final model</a> uses an optimised MLP classifier as well as a Random Forest for comparison.

The learning curve shows how the accuracy increases as the size of the training data increases. For this the sklearn method `learning_curve()` was used, which can automatically plot a learning curve. The test sizes were the following: `[10, 20, 30, 40, 50, 60, 70, 80, 90, 100, 200, 300, 400, 500, 600, 700, 800, 900, 1000, 2000, 3000, 4000, 5000]`. On the graph you can also see the comparison with two baselines: Naive Bayes and Random Class classifier, as mentioned above. The results show a steady increase in accuracy up to around 1000, from which point the increase in accuracy slows, however is still present.

<img src="images/learning_curve_w_mlp_fade.png" width=600>

1000 data points, roughly 63 data points per class, is evidently enough to learn the most important features (as we can see <a href="https://gitlab.cl.uni-heidelberg.de/innes/exp-ml-1/-/blob/master/project/README.md#feature-importance">above</a>, in most cases this means particular letters). Interesting to note is also that Decision Tree and Random Forest both have 100% accuracy on the training set, whereas Naive Bayes only has around 90%. 

In order to obtain a more reliable measure of how well the model works, here is a breakdown of the precision, recall and F1-score per class using a Random Forest classifier. This was created using the `classification_report()` function in `sklearn.metrics` (source code <a href="evaluation/plot_classification_report.py">here</a>).

```
              precision    recall  f1-score   support

          be       0.99      1.00      1.00       121
          bg       0.99      1.00      0.99       132
          cs       0.99      0.98      0.99       126
          cu       1.00      1.00      1.00       121
         dsb       0.96      0.98      0.97       126
          hr       0.96      0.92      0.94       114
         hsb       0.98      0.95      0.97       108
          mk       0.98      1.00      0.99       118
          pl       0.99      0.99      0.99       123
          ru       0.94      0.97      0.96       104
         rue       0.97      0.95      0.96       140
          sk       0.96      0.99      0.97        87
          sl       0.93      0.97      0.95       115
          sr       1.00      0.98      0.99       125
         szl       1.00      0.98      0.99       123
          uk       0.99      0.98      0.98       127

    accuracy                           0.98      1910
   macro avg       0.98      0.98      0.98      1910
weighted avg       0.98      0.98      0.98      1910
```

As is clear from the F1-scores in the above table, the model is accurate and precise for all classes. It could of course have been the case that one class is never predicted, and the accuracy could still be well over 90%, however, using the F-scores, which are a weighted average of accuracy and precision (in this case F1 where $`=2\cdot\frac{precision \cdot recall}{precision + recall}`$), we know that this is not the case here.

However, the above metrics are generated by testing the model on the Wikipedia corpus. This would logically lead to a higher accuracy than on "real-world" texts, since there are certain words which appear more often on Wikipedia pages due to the nature of the topics (see the discussion on <a href= "https://gitlab.cl.uni-heidelberg.de/innes/exp-ml-1/-/blob/master/project/README.md#features">features</a>). Since testing the algorithm on real-world data requires annotation, which is beyond the realms of the resources of this project, it is impossible to provide a numerical value for the accuracy. However, using Article 1 of the UN Declaration of Human Rights, all the languages for which a translation was available were tested.

> All human beings are born free and equal in dignity and rights. They are endowed with reason and conscience and should act towards one another in a spirit of brotherhood.

> Все люди рождаются свободными и равными в своем достоинстве и правах. Они наделены разумом и совестью и должны поступать в отношении друг друга в духе братства.

Article 1 is comparatively relatively short - the length of the English text is `30` words, in Russian `26` - so the model works less well, both algorithms classifying 13 of the 16 languages correctly. Using the somewhat longer Article 2 - `79` words in Russian - the accuracy increased. The Article 1 translations were retrieved from <a href="https://omniglot.com/udhr/index.htm">Omniglot</a>, and the Article 2 texts from the <a href="https://www.ohchr.org/EN/UDHR/Pages/SearchByLang.aspx">United Nations official translations</a>. 

| Supported languages: | Article 1 (MLP) | Article 1 (RF) | Article 2 (MLP) | Article 2 (RF) |
| --- | --- | --- | --- | --- |
| Russian | :x: | :white_check_mark: | :white_check_mark: | :white_check_mark: |
| Polish | :white_check_mark: | :white_check_mark: | :white_check_mark: | :white_check_mark: |
| Ukrainian | :white_check_mark: | :white_check_mark: | :white_check_mark: | :white_check_mark: |
| Slovenian | :x: | :white_check_mark: | :white_check_mark: | :x: |
| Slovakian | :white_check_mark: | :white_check_mark: | :white_check_mark: | :white_check_mark: |
| Bulgarian | :white_check_mark: | :white_check_mark: | :white_check_mark: | :white_check_mark: |
| Belarussian | :x: | :white_check_mark: | :white_check_mark: | :white_check_mark: |
| Czech | :white_check_mark: | :white_check_mark: | :white_check_mark: | :white_check_mark: |
| Macedonian | :white_check_mark: | :x: | :x: | :white_check_mark: |
| Serbo-Croatian: Cyrillic | :white_check_mark: | :x: | :white_check_mark: | :x: |
| Serbo-Croatian: Latin | :white_check_mark: | :x: | :white_check_mark: | :white_check_mark: |
| Silesian | :white_check_mark: | :white_check_mark: | N/A | N/A |
| Upper Sorbian | :white_check_mark: | :white_check_mark: | N/A | N/A |
| Lower Sorbian | :white_check_mark: | :white_check_mark: | N/A | N/A |
| Rusyn | :white_check_mark: | :white_check_mark: | N/A | N/A |
| Old Church Slavonic: Cyrillic | :white_check_mark: | :white_check_mark: | N/A | N/A |

To further test the languages for which Article 2 was not available, paragraphs were taken from other sources, which was sometimes tricky as the 5 languages have little to no internet presence. Below is an example of the first two lines of the Lord's Prayer, correctly classified as Silesian, and the output.

```
Enter text to be classified: Ôjcze nŏsz, kery jeżeś we niebie, bydź poświyncōne miano Twoje.

Random Forest prediction:
Silesian

Would you like find out more about this language? (y/n) y
Here is the Wikipedia page for the language: https://en.wikipedia.org/wiki/Silesian_language

Would you like the see the probabilites for other languages? (y/n) y
Belarusian: 0.0
Bulgarian: 0.0
Czech: 0.01
Old Church Slavonic: 0.02
Lower Sorbian: 0.06
Serbo-Croatian: Latin: 0.0
Upper Sorbian: 0.06
Macedonian: 0.0
Polish: 0.12
Russian: 0.0
Rusyn: 0.0
Slovak: 0.01
Slovene: 0.01
Serbo-Croat: Cyrillic: 0.03
Silesian: 0.68
Ukrainian: 0.0

Would you like to enter another text? (y/n)
```
The next highest probability was Polish, which is to be expected due to the similarity of the two languages (<a href="https://cadmus.eui.eu/bitstream/handle/1814/1351/HEC03-01.pdf">some regard Silesian as a dialect of Polish</a>), and it is mostly other languages written in the Latin script, which are predicted, also unsurprising.

It is therefore clear that, while no precise accuracy can be given for real-world data, the model works very reliably, given the document is greater than a particular length (around 50 words). One possible way of calculating more reliable metrics on this point without using annotation could be to use several single-language corpora and use that data to calculate metrics.

## Future work
This model was purely focussed on Slavic languages however could easily be extended to other language groups which have a Wikipedia. In order to do this, one would have to update the alphabet, and it might be necessary to change the number of n-most common words in the BoW model. A more complex model could also use N-grams to capture morphological information in a more sophisticated way.

<img src="images/birch-forest-crop.jpg" width=2000/>

# Technical details

## System Requirements
Computer<br>
Python 3.9 (or similar)<br>
Scikit-learn<br>

## Installation
For details on how to download and use the model, see the <a href="https://gitlab.cl.uni-heidelberg.de/innes/exp-ml-1/-/blob/master/README.md">introductory README for the seminar</a>.

## Author
Samuel Innes: dd257@stud.uni-heidelberg.de

## References
[^1]: Image source: http://strongwood.lv/slide-page/final_full_width/
[^2]: https://www.britannica.com/topic/Slavic-languages
[^3]: Image source: https://en.wikipedia.org/wiki/Slavic_languages
[^4]: Image source: https://commons.wikimedia.org/w/index.php?curid=22112928
[^5]: https://www.britannica.com/topic/Slavic-languages/Linguistic-characteristics
[^6]: Image source: https://web.stanford.edu/~jurafsky/slp3/4.pdf
