# this module tries out different values for hyperparameters in Decision Tree
# and Random Forest classifiers to optimise performance.
import numpy as np
import re
import os

from sklearn.model_selection import train_test_split
from sklearn.ensemble import RandomForestClassifier
from sklearn.metrics import accuracy_score
from sklearn.model_selection import cross_val_score

import matplotlib.pyplot as plt
from sklearn.tree import DecisionTreeClassifier

CWD = os.getcwd()

with open(CWD + "/project/data/train_data.txt", "r") as f:
    data = f.read().strip().split("\t")

# extract data from train_data.txt
X_train_and_dev = []
y_train_and_dev = []
for dp in data:
    with_bracket = re.search("\([^a-zA-Z]+\)", dp).group()[1:]
    no_bracket = with_bracket.strip("(").strip(")").split(", ")
    X_train_and_dev.append(np.array(no_bracket, dtype=np.int32))
    y_train_and_dev.append(re.search("[a-zA-Z]+", dp).group())

# create train/dev split - could be used for manual evaluation
# in this case, however, I decided to use Cross-Validation
X_train, X_dev, y_train, y_dev = \
    train_test_split(X_train_and_dev, y_train_and_dev, test_size=0.2, random_state=0) # then create dev set from train set

# this is the baseline of the default value for max_depth
rf = RandomForestClassifier(random_state=0)
rf.fit(X_train, y_train)
predictions = rf.predict(X_dev)
baseline_accuracy = accuracy_score(y_dev, predictions)
print("baseline: " + str(baseline_accuracy))

# try out different max_depth scores
accuracy_scores = []
i_s = []
for i in range(1, 31):  
    rf = RandomForestClassifier(max_depth=i, random_state=0)
    cvs = cross_val_score(rf, X_train_and_dev, y_train_and_dev, cv=5)                       # create 5 folds
    accuracy = sum(cvs)/len(cvs)                                # calculate mean accuracy
    accuracy_scores.append(accuracy)
    i_s.append(i)
    print(str(i) + ": " + str(accuracy))

# plot graph
plt.plot(i_s, accuracy_scores, label="Accuracy by changing max_depth")
plt.plot(i_s, [baseline_accuracy]*len(i_s), label = "Baseline accuracy")
plt.xlabel('max_depth')
plt.ylabel('accuracy')
plt.legend(loc='best')
plt.show()



# investigate influence of splitter type
rf = DecisionTreeClassifier(random_state=0, splitter="best")
cvs = cross_val_score(rf, X_train_and_dev, y_train_and_dev, cv=5)                       # create 5 folds
accuracy = sum(cvs)/len(cvs)                                # calculate mean accuracy
print("Best (default): " + str(accuracy))

rf = DecisionTreeClassifier(random_state=0, splitter="random")
cvs = cross_val_score(rf, X_train_and_dev, y_train_and_dev, cv=5)                       # create 5 folds
accuracy = sum(cvs)/len(cvs)                                # calculate mean accuracy
print("Random: " + str(accuracy))


# this is the baseline of the default value for max_depth
rf = RandomForestClassifier(random_state=0)
rf.fit(X_train, y_train)
predictions = rf.predict(X_dev)
baseline_accuracy = accuracy_score(y_dev, predictions)
print("baseline: " + str(baseline_accuracy))

# try out different max_depth scores
accuracy_scores = []
i_s = []
for i in range(2, 100):  
    rf = RandomForestClassifier(min_samples_split=i, random_state=0)
    cvs = cross_val_score(rf, X_train_and_dev, y_train_and_dev, cv=5)                       # create 5 folds
    accuracy = sum(cvs)/len(cvs)                                # calculate mean accuracy
    accuracy_scores.append(accuracy)
    i_s.append(i)
    print(str(i) + ": " + str(accuracy))

# plot graph
plt.plot(i_s, accuracy_scores, label="Accuracy by changing min_samples_split")
plt.plot(i_s, [baseline_accuracy]*len(i_s), label = "Baseline accuracy")
plt.xlabel('min_samples_split')
plt.ylabel('accuracy')
plt.legend(loc='best')
plt.show()