# this module tries out different values for hyperparameters in the Multi-Layer Perceptron
# classifier to optimise performance.

import os 
import re
import numpy as np

from sklearn.neural_network import MLPClassifier
from sklearn.model_selection import GridSearchCV

CWD = os.getcwd()

with open(CWD + "/project/data/train_data.txt", "r") as f:
    data = f.read().strip().split("\t")

# extract data from train_data.txt
X_train_and_dev = []
y_train_and_dev = []
for dp in data:
    with_bracket = re.search("\([^a-zA-Z]+\)", dp).group()[1:]
    no_bracket = with_bracket.strip("(").strip(")").split(", ")
    X_train_and_dev.append(np.array(no_bracket, dtype=np.int32))
    y_train_and_dev.append(re.search("[a-zA-Z]+", dp).group())

parameter_space = {
    'hidden_layer_sizes': [(50,100,50), (100,)],
    'activation': ['tanh', 'relu'],
    'solver': ['lbfgs', 'adam'],
    'alpha': [0.0001, 0.05],
}

nn = MLPClassifier(max_iter=100)

clf = GridSearchCV(nn, parameter_space, n_jobs=-1, cv=5)
clf.fit(X_train_and_dev, y_train_and_dev)

# Best set of parameters
print('Optimal parameters found:\n', clf.best_params_)

# All results
means = clf.cv_results_['mean_test_score']
stds = clf.cv_results_['std_test_score']
for mean, std, params in zip(means, stds, clf.cv_results_['params']):
    print("%0.3f (+/-%0.03f) for %r" % (mean, std * 2, params))