# this module tries out different values for hyperparameters in Naive Bayes
# classifier to optimise performance.

import os 
import re
from matplotlib import pyplot as plt

import numpy as np

from sklearn.metrics import accuracy_score
from sklearn.model_selection import cross_val_score, train_test_split
from sklearn.naive_bayes import MultinomialNB

CWD = os.getcwd()

with open(CWD + "/project/data/train_data.txt", "r") as f:
    data = f.read().strip().split("\t")

# extract data from train_data.txt
X_train_and_dev = []
y_train_and_dev = []
for dp in data:
    with_bracket = re.search("\([^a-zA-Z]+\)", dp).group()[1:]
    no_bracket = with_bracket.strip("(").strip(")").split(", ")
    X_train_and_dev.append(np.array(no_bracket, dtype=np.int32))
    y_train_and_dev.append(re.search("[a-zA-Z]+", dp).group())

# create train/dev split - could be used for manual evaluation
# in this case, however, I decided to use Cross-Validation
X_train, X_dev, y_train, y_dev = \
    train_test_split(X_train_and_dev, y_train_and_dev, test_size=0.2, random_state=0) # then create dev set from train set

nb = MultinomialNB()
cvs = cross_val_score(nb, X_train_and_dev, y_train_and_dev, cv=5)                       # create 5 folds
baseline_accuracy = sum(cvs)/len(cvs)                                # calculate mean accuracy
print("fit_prior=False: " + str(baseline_accuracy))

# create non-optimised baseline
accuracy_scores = []
i_s = []
for i in [0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1]:
    nb = MultinomialNB(alpha=i)
    cvs = cross_val_score(nb, X_train_and_dev, y_train_and_dev, cv=5)                       # create 5 folds
    accuracy = sum(cvs)/len(cvs)                                # calculate mean accuracy
    accuracy_scores.append(accuracy)
    i_s.append(i)
    print(str(i) + ": " + str(accuracy))

# plot graph
plt.plot(i_s, accuracy_scores, label="Accuracy by changing alpha (smoothing)")
plt.plot(i_s, [baseline_accuracy]*len(i_s), label = "Baseline accuracy")
plt.xlabel('alpha')
plt.ylabel('accuracy')
plt.legend(loc='best')
plt.show()

# investigate effect of fit_prior on accuracy
nb = MultinomialNB(fit_prior=False)
cvs = cross_val_score(nb, X_train_and_dev, y_train_and_dev, cv=5)                       # create 5 folds
accuracy = sum(cvs)/len(cvs)                                # calculate mean accuracy
print("fit_prior=False: " + str(accuracy))

nb = MultinomialNB(fit_prior=True)
cvs = cross_val_score(nb, X_train_and_dev, y_train_and_dev, cv=5)                       # create 5 folds
accuracy = sum(cvs)/len(cvs)                                # calculate mean accuracy
print("fit_prior=True: " + str(accuracy))