# this module investigates the importance of certain groups of feature for the model

import numpy as np
import pandas as pd
import re
import os
import matplotlib.pyplot as plt

from sklearn.model_selection import cross_val_score
from sklearn.ensemble import RandomForestClassifier
from sklearn.naive_bayes import MultinomialNB
from sklearn.tree import DecisionTreeClassifier
from sklearn.neural_network import MLPClassifier

CWD = os.getcwd()

with open(CWD + "/project/data/train_data.txt", "r") as f:
    data = f.read().strip().split("\t")

# extract data from train_data.txt
X_train_and_dev = []
y_train_and_dev = []
for dp in data:
    with_bracket = re.search("\([^a-zA-Z]+\)", dp).group()[1:]
    no_bracket = with_bracket.strip("(").strip(")").split(", ")
    X_train_and_dev.append(np.array(no_bracket, dtype=np.int32))
    y_train_and_dev.append(re.search("[a-zA-Z]+", dp).group())

# isolate just the words
X_train_words = []
for dp in X_train_and_dev:
    X_train_words.append(dp[:-129])

# isolate just the characters
X_train_char = []
for dp in X_train_and_dev:
    X_train_char.append(dp[-129:])


# hardcode feature_list
feature_list = ['в', 'и', 'на', 'с', 'года', 'по', 'из', 'что', 'а', 'году', 'w', 'i', 'z', 'na', 'do', 'się', \
    'roku', 'a', 'od', 'po', 'в', 'у', 'на', 'з', 'і', 'та', 'до', 'року', 'за', 'що', 'je', 'in', 'v', 'na', \
    'so', 'se', 'za', 'ki', 'leta', 'z', 'v', 'a', 'na', 'je', 'sa', 'z', 's', 'roku', 'm', 'ako', 'је', 'у', 'и', \
    'на', 'се', 'су', 'од', 'за', 'да', 'године', 'на', 'и', 'в', 'е', 'от', 'се', 'за', 'с', 'г', 'да', 'і', 'у', \
    'ў', 'з', 'на', 'года', 'да', 'па', 'а', 'г', 'на', 'во', 'и', 'од', 'се', 'за', 'е', 'со', 'да', 'година', 'v', \
    'a', 'na', 'se', 'je', 's', 'z', 'do', 've', 'byl', 'je', 'i', 'u', 'na', 'se', 'su', 'od', 'za', 'a', 's', 'we', \
    'je', 'i', 'do', 'co', 'przipisy', 'nŏleży', 'zorty', 'go', 'ôpisoł', 'a', 'jo', 'w', 'na', 'se', 'z', 'wót', \
    'su', 'do', 'až', 'a', 'w', 'je', 'na', 'wot', 'z', 'so', 'do', 'k', 'za', 'и', 'ѥстъ', 'лѣта', 'бѣ', 'градъ', \
    'жє', 'въ', 'ꙁьри', 'такождє', 'людии', 'в', 'и', 'на', 'ся', 'і', 'з', 'є', 'до', 'року', 'а', 'а', 'б', 'в', \
    'г', 'д', 'е', 'ё', 'ж', 'з', 'и', 'й', 'к', 'л', 'м', 'н', 'о', 'п', 'р', 'с', 'т', 'у', 'ф', 'х', 'ц', 'ч', \
    'ш', 'щ', 'ъ', 'ы', 'ь', 'э', 'ю', 'я', 'a', 'ą', 'b', 'c', 'ć', 'd', 'e', 'ę', 'f', 'g', 'h', 'i', 'j', 'k', \
    'l', 'ł', 'm', 'n', 'ń', 'o', 'ó', 'p', 'q', 'r', 's', 'ś', 't', 'u', 'v', 'w', 'x', 'y', 'z', 'ź', 'ż', 'ґ', \
    'є', 'і', 'ї', 'č', 'š', 'ž', 'á', 'ä', 'ď', 'dz', 'dž', 'é', 'ch', 'í', 'ĺ', 'ľ', 'ň', 'ô', 'ŕ', 'ť', 'ú', 'ý', \
    'ђ', 'ј', 'љ', 'њ', 'ћ', 'џ', 'ў', 'ѓ', 'ѕ', 'ќ', 'ě', 'ř', 'ů', 'ã', 'ŏ', 'ō', 'õ', 'ꙃ', 'ꙁ', 'ꙉ', 'оу', 'ꙋ', \
    'ѡ', 'ѿ', 'ꙑ', 'ъи', 'ѣ', 'ꙗ', 'ѥ', 'ѧ', 'ѩ', 'ѫ', 'ѭ', 'ѯ', 'ѱ', 'ѳ', 'ѵ', 'đ']
# investigate influence of word and character vectors
# trained on BoW

nn = MLPClassifier(activation='tanh')
cvs = cross_val_score(nn, X_train_words, y_train_and_dev, cv=5)
cvs = sum(cvs)/len(cvs)
print("MLP")
print(cvs)

rfcf = RandomForestClassifier(random_state=0)
cvs = cross_val_score(rfcf, X_train_words, y_train_and_dev, cv=5)
cvs = sum(cvs)/len(cvs)
print("Random Forest")
print(cvs)

rfcf = DecisionTreeClassifier(random_state=0)
cvs = cross_val_score(rfcf, X_train_words, y_train_and_dev, cv=5)
cvs = sum(cvs)/len(cvs)
print("Decision Tree")
print(cvs)

rfcf = MultinomialNB()
cvs = cross_val_score(rfcf, X_train_words, y_train_and_dev, cv=5)
cvs = sum(cvs)/len(cvs)
print("Naive Bayes")
print(cvs)

# trained on characters

cvs = cross_val_score(nn, X_train_char, y_train_and_dev, cv=5)
cvs = sum(cvs)/len(cvs)
print("MLP")
print(cvs)

cvs = cross_val_score(rfcf, X_train_char, y_train_and_dev, cv=5)
cvs = sum(cvs)/len(cvs)
print("Random Forest")
print(cvs)

cvs = cross_val_score(rfcf, X_train_char, y_train_and_dev, cv=5)
cvs = sum(cvs)/len(cvs)
print("Decision Tree")
print(cvs)

cvs = cross_val_score(rfcf, X_train_char, y_train_and_dev, cv=5)
cvs = sum(cvs)/len(cvs)
print("Naive Bayes")
print(cvs)

# combined

cvs = cross_val_score(nn, X_train_and_dev, y_train_and_dev, cv=5)
cvs = sum(cvs)/len(cvs)
print("MLP")
print(cvs)

cvs = cross_val_score(rfcf, X_train_and_dev, y_train_and_dev, cv=5)
cvs = sum(cvs)/len(cvs)
print("Random Forest")
print(cvs)

cvs = cross_val_score(rfcf, X_train_and_dev, y_train_and_dev, cv=5)
cvs = sum(cvs)/len(cvs)
print("Decision Tree")
print(cvs)

cvs = cross_val_score(rfcf, X_train_and_dev, y_train_and_dev, cv=5)
cvs = sum(cvs)/len(cvs)
print("Naive Bayes")
print(cvs)


# code from https://scikit-learn.org/stable/auto_examples/ensemble/plot_forest_importances.html
rf = RandomForestClassifier(random_state=0)
rf.fit(X_train_and_dev, y_train_and_dev)
importances = rf.feature_importances_
std = np.std([tree.feature_importances_ for tree in rf.estimators_], axis=0)

forest_importances = pd.Series(importances, index=feature_list)

fig, ax = plt.subplots()
forest_importances.plot.bar(yerr=std, ax=ax)
ax.set_title("Feature importances using MDI")
ax.set_ylabel("Mean decrease in impurity")
ax.set_xlabel("Features")
fig.tight_layout()
plt.tick_params(labelbottom=False) # to remove x-axis labels
plt.show()

