# this module splits original data into training and test data, writing them into separate files

import numpy as np
import re
import os

from sklearn.utils import shuffle
from sklearn.model_selection import train_test_split

CWD = os.getcwd()

with open(CWD + "/project/data/data.txt", "r") as f:
    data = f.read().strip().split("\t")

# extract data from data.txt
X = []
y = []
for dp in data:
    with_bracket = re.search("\([^a-zA-Z]+\)", dp).group()[1:]
    no_bracket = with_bracket.strip("(").strip(")").split(", ")
    # convert X to np.array (to remove "") then to tuple to fit the formatting of data.txt
    X.append(tuple(np.array(no_bracket, dtype=np.int32).tolist()))
    y.append(re.search("[a-zA-Z]+", dp).group())

# separate into train/dev/test sets with proportions 60/15/25
# shuffle data so that all languages can be learned
shuffled_X, shuffled_y = shuffle(X, y, random_state=0)

X_train, X_test, y_train, y_test = \
    train_test_split(shuffled_X, shuffled_y, test_size=0.25, random_state=0) # separate test data

# write training data to file train_data.txt
with open(CWD + "/project/data/train_data.txt", "w+", encoding="utf-8") as f:
    for dp in zip(X_train, y_train):
        f.write(str(dp)+"\t")

# write test data to file test_data.txt
with open(CWD + "/project/data/test_data.txt", "w+", encoding="utf-8") as f:
    for dp in zip(X_test, y_test):
        f.write(str(dp)+"\t")