# this module prints cross validation for three different classifiers

import re
import os
import numpy as np

from sklearn.ensemble import RandomForestClassifier
from sklearn.tree import DecisionTreeClassifier
from sklearn.naive_bayes import MultinomialNB
from sklearn.neural_network import MLPClassifier
from sklearn.model_selection import cross_val_score


CWD = os.getcwd()
LANG_LIST = ["ru", "pl", "uk", "sl", "sk", "sr", "bg", "be", "mk", "cs", "hr", "szl", "dsb", "hsb", "cu", "rue"]

# read training data
with open(CWD + "/project/data/train_data.txt", "r") as f:
    train_data = f.read().strip().split("\t")

# extract data from train_data.txt
X_train = []
y_train = []
for dp in train_data:
    with_bracket = re.search("\([^a-zA-Z]+\)", dp).group()[1:]
    no_bracket = with_bracket.strip("(").strip(")").split(", ")
    X_train.append(np.array(no_bracket, dtype=np.int32))
    y_train.append(re.search("[a-zA-Z]+", dp).group())

nn = MLPClassifier(activation="tanh")
cvs = cross_val_score(nn, X_train, y_train, cv=5)
print("Multi-Layer Perceptron")
print(cvs)
print("Average:")
print(sum(cvs)/len(cvs))

rfcf = RandomForestClassifier(random_state=0)
cvs = cross_val_score(rfcf, X_train, y_train, cv=5)
print("Random Forest")
print(cvs)
print("Average:")
print(sum(cvs)/len(cvs))

dtcf = DecisionTreeClassifier()
cvs = cross_val_score(dtcf, X_train, y_train, cv=5)
print("Decision Tree")
print(cvs)
print("Average:")
print(sum(cvs)/len(cvs))

nb = MultinomialNB()
cvs = cross_val_score(nb, X_train, y_train, cv=5)
print("Naive Bayes")
print(cvs)
print("Average:")
print(sum(cvs)/len(cvs))
