# this module plots a learning curve with 4 different classifiers

import re
import os
import numpy as np
import matplotlib.pyplot as plt
import dummy_classifiers as dc

from sklearn.utils import shuffle
from sklearn.model_selection import learning_curve
from sklearn.ensemble import RandomForestClassifier
from sklearn.tree import DecisionTreeClassifier
from sklearn.dummy import DummyClassifier
from sklearn.naive_bayes import MultinomialNB
from sklearn.neural_network import MLPClassifier

CWD = os.getcwd()
LANG_LIST = ["ru", "pl", "uk", "sl", "sk", "sr", "bg", "be", "mk", "cs", "hr", "szl", "dsb", "hsb", "cu", "rue"]

with open(CWD + "/project/data/data.txt", "r") as f:
    data = f.read().strip().split("\t")

# extract data from data.txt
X = []
y = []
for dp in data:
    with_bracket = re.search("\([^a-zA-Z]+\)", dp).group()[1:]
    no_bracket = with_bracket.strip("(").strip(")").split(", ")
    X.append(np.array(no_bracket, dtype=np.int32))
    y.append(re.search("[a-zA-Z]+", dp).group())

# shuffle data to remove bias of order of data
X, y = shuffle(X, y, random_state=0)

# create train sizes
train_sizes = []
for i in range(10, 101, 10): # add 10, 20 ...
    train_sizes.append(i)

for i in range(200, 1001, 100): # add 100, 200 ...
    train_sizes.append(i)

for i in range(2000, 6000, 1000): # add 1000, 2000 ...
    train_sizes.append(i)
print(train_sizes)

def plot_learning_curve(estimator, train_sizes, label, spec_color):
    """
    Plots a learning curve for a given estimator (classifier).

            Parameters:
                    estimator: classifier
                    train_sizes (list): list of training set sizes to plot graph over
                    label (string): name of classifier on end graph

            Returns:
                    None
    """
    train_sizes, train_scores, test_scores = learning_curve(estimator, X, y, train_sizes = train_sizes, shuffle=True)
    test_score_means = []
    for test_score in test_scores:
        test_score_means.append(np.mean(test_score))

    train_score_means = []
    for train_score in train_scores:
        train_score_means.append(np.mean(train_score))


    plt.plot(train_sizes, test_score_means, label=label + " (test)", linewidth=0.7, color=spec_color)
    plt.plot(train_sizes, train_score_means, label=label + " (train)", linewidth=0.7, color=spec_color, alpha=0.5)

    
rfcf = RandomForestClassifier(random_state=0)
plot_learning_curve(rfcf, train_sizes, "Random Forest", "tab:green")
print("Random Forest finished")

clf = DecisionTreeClassifier(random_state=0)
plot_learning_curve(clf, train_sizes, "Decision Tree", "tab:blue")
print("Decision Tree finished")

nn = MLPClassifier()
plot_learning_curve(nn, train_sizes, "Neural Network", "tab:orange")
print("Neural Network finished")

nb = MultinomialNB()
plot_learning_curve(nb, train_sizes, "Naive Bayes (baseline)", "tab:red")
print("Naive Bayes finished")

dc = DummyClassifier(random_state=0, strategy="stratified")
plot_learning_curve(dc, train_sizes, "Random Class (baseline)", "tab:gray")
print("Random Class finished")

# format graph
plt.title("Learning curve")
plt.xlabel('train_size')
plt.ylabel('accuracy')
plt.legend(loc='best')

leg = plt.legend()
leg_lines = leg.get_lines()
plt.setp(leg_lines, linewidth=1)

plt.show()