<img src="images/birch-forest.jpg" width=2000/>[^1]

# Slavic language classifier
This project takes as input a text in a particular Slavic language and returns which language it is, including information about the linguistic classification of that language.

### Contents:
- <a href="https://gitlab.cl.uni-heidelberg.de/innes/exp-ml-1/-/blob/master/project/README.md#background">Background</a>
    - <a href="https://gitlab.cl.uni-heidelberg.de/innes/exp-ml-1/-/blob/master/project/README.md#motivation">Motivation</a>
    - <a href="https://gitlab.cl.uni-heidelberg.de/innes/exp-ml-1/-/blob/master/project/README.md#similar-projects">Similar Projects</a>
    - <a href="https://gitlab.cl.uni-heidelberg.de/innes/exp-ml-1/-/blob/master/project/README.md#the-slavic-language-family">The Slavic language family</a>
- <a href="">Model</a>
    - <a href="https://gitlab.cl.uni-heidelberg.de/innes/exp-ml-1/-/blob/master/project/README.md#supported-languages">Supported languages</a>
    - <a href="https://gitlab.cl.uni-heidelberg.de/innes/exp-ml-1/-/blob/master/project/README.md#training-corpus-and-features">Training corpus and features</a>
    - <a href="https://gitlab.cl.uni-heidelberg.de/innes/exp-ml-1/-/blob/master/project/README.md#which-algorithms-could-be-used-to-solve-this-problem">Which algorithms could be used to solve this problem?</a>
- <a href="https://gitlab.cl.uni-heidelberg.de/innes/exp-ml-1/-/blob/master/project/README.md#performance">Performance</a>
    - <a href="https://gitlab.cl.uni-heidelberg.de/innes/exp-ml-1/-/blob/master/project/README.md#how-can-the-performance-be-measured">How can the performance be measured?</a>
    - <a href="https://gitlab.cl.uni-heidelberg.de/innes/exp-ml-1/-/blob/master/project/README.md#and-what-performance-is-expected">And what performance is expected?</a>
- <a href="https://gitlab.cl.uni-heidelberg.de/innes/exp-ml-1/-/blob/master/project/README.md#technical-details">Technical details</a>
    - <a href = "https://gitlab.cl.uni-heidelberg.de/innes/exp-ml-1/-/blob/master/project/README.md#system-requirements">System Requirements</a>
    - <a href = "https://gitlab.cl.uni-heidelberg.de/innes/exp-ml-1/-/blob/master/project/README.md#installation">Installation</a>
    - <a href = "https://gitlab.cl.uni-heidelberg.de/innes/exp-ml-1/-/blob/master/project/README.md#todos">TODOs</a>
    - <a href = "https://gitlab.cl.uni-heidelberg.de/innes/exp-ml-1/-/blob/master/project/README.md#common-problems">Common Problems</a>
    - <a href = "https://gitlab.cl.uni-heidelberg.de/innes/exp-ml-1/-/blob/master/project/README.md#author">Author</a>
- <a href="https://gitlab.cl.uni-heidelberg.de/innes/exp-ml-1/-/blob/master/project/README.md#references">References</a>

# Background

## Motivation
Although many language detectors already exist, there are, as far as I can tell, none which focus specifically on the Slavic language family and are thus able to detect less widely-spoken Slavic languages such as Upper Sorbian or Rusyn. However, using the fact that Wikipedia is available in many of this smaller languages, a classifier can be trained, as is explained below, to detect these minority languages. Such a classifier can help towards preservation of these languages and can contribute to their digital presence, thus encouraging speakers to use them.

## Similar projects
The most obvious and popular example for language recognition is <a href="https://translate.google.co.uk/">Google Translate</a>. The amount of languages offered is ever growing, however of the 14 supported languages listed below, Google Translate can only detect 10. <a href="https://www.deepl.com/translator">DeepL</a> is another Machine Translation tool that has become popular in recent years, however has an even smaller selection of languages. 

There are other specialist websites offering language classification, such as <a href="https://www.dcode.fr/language-recognition">dcode</a> or <a href="whatlanguageisthis.com">whatlanguageisthis.com</a>, however these also have no support for lesser spoken Slavic languages, with an even smaller range than Google Translate.

## The Slavic language family
The Encyclopedia Britannica describes the Slavic language family thus:
><i>Slavic languages, also called Slavonic languages, group of Indo-European languages spoken in most of eastern Europe, much of the Balkans, parts of central Europe, and the northern part of Asia. The Slavic languages, spoken by some 315 million people at the turn of the 21st century, are most closely related to the languages of the Baltic group (Lithuanian, Latvian, and the now-extinct Old Prussian), but they share certain linguistic innovations with the other eastern Indo-European language groups (such as Indo-Iranian and Armenian) as well.<i>[^2]

<img src="images/Slavic_europe.jpg" width=400>

[^3]

# Model

## Supported languages
The list of wikipedia sites in Slavic languages is a follows:
Russian, Polish, Ukranian, Slovenian, Slovakian, Bulgarian, Belarussian, Czech, Croatian, Macedonian, Serbian*, Serbo-Croatian*, Bosnian*, Silesian, Lower Sorbian, Rusyn, Old Church Slavonic (in Cyrillic and Glagolitic script)*.

However, due to technical reasons, I have to make the following adjustments to the list:
<ul>
  <li>Due to the linguistic similarity of the Serbo-Croatian language group and in order to avoid inadvertent political statements, I have decided to replace the individual languages/dialects of Serbian, Croatian and Bosnian with two categories: Serbo-Croatian: Cyrillic and Serbo-Croatian: Latin.
  <li>Old Church Slavonic wikipedia is also available in two alphabets: Cyrillic and Glagolitic. The corpus size is small, however the Cyrillic version uses letters not found in any other language, and the Glagolitic version is unique to Old Church Slavonic (with the exception of a few Croatian dialects). Hence I have created two labels for the language: Old Church Slavonic: Cyrillic and Old Church Slavonic: Glagolitic.
</ul>

Therefore the list labels in the model is the following (in rough order of wikipedia size):
| Supported languages: | 
| --- | 
Russian
Polish
Ukranian
Slovenian
Slovakian
Bulgarian
Belarussian
Czech
Macedonian
Serbo-Croatian: Cyrillic
Serbo-Croatian: Latin
Silesian
Lower Sorbian
Rusyn
Old Church Slavonic: Cyrillic
Old Church Slavonic: Glagolitic

## Training corpus and features
Using a webscraping to extract text from wikipedia, I can create reliable text corpora, even for lesser spoken Slavic languages. For a more indepth discussion of the use of Wikipedia as a corpus, see <a href="PasternackRo08.pdf">Pasternak & Roth 2008</a>. The smallest wikipedia in a Slavic language is Old Church Slavonic with 846 articles. Although short, this should give enough tokens to create a bag of words model. The features would therefore be the n most common words in the language (i.e. the stop words). I would then make a feature vector with all the stop words in all languages. Words that are spelled the same in several languages of course only appear once.

Some example document vectors could look something like this:
<table>
  <tr>
    <td colspan="4"><b>Russian</b></td>
    <td colspan="4"><b>Bulgarian</b></td>
    <td>...</td>
    <td colspan="4"><b>Polish</b></td>
  </tr>
  <tr>
    <!--Russian words-->
    <td>и</td>
    <td>в</td>
    <td>не</td>
    <td>...</td>
    <!--Bulgarian words-->
    <td>да</td>
    <td>се</td>
    <td>на</td>
    <td>...</td>
    <!--...-->
    <td>...</td>
    <!--Polish words-->
    <td>nie</td>
    <td>to</td>
    <td>się</td>
    <td>...</td>
  </tr>
  <tr>
    <!--Russian words-->
    <td>0</td>
    <td>0</td>
    <td>1</td>
    <td>...</td>
    <!--Bulgarian words-->
    <td>5</td>
    <td>3</td>
    <td>5</td>
    <td>...</td>
    <!--...-->
    <td>...</td>
    <!--Polish words-->
    <td>0</td>
    <td>0</td>
    <td>0</td>
    <td>...</td>
  </tr>
  <tr>
    <!--Russian words-->
    <td>2</td>
    <td>0</td>
    <td>0</td>
    <td>...</td>
    <!--Bulgarian words-->
    <td>1</td>
    <td>0</td>
    <td>0</td>
    <td>...</td>
    <!--...-->
    <td>...</td>
    <!--Polish words-->
    <td>0</td>
    <td>0</td>
    <td>0</td>
    <td>...</td>
  </tr>
  <tr>
    <!--Russian words-->
    <td>7</td>
    <td>9</td>
    <td>8</td>
    <td>...</td>
    <!--Bulgarian words-->
    <td>1</td>
    <td>2</td>
    <td>5</td>
    <td>...</td>
    <!--...-->
    <td>...</td>
    <!--Polish words-->
    <td>0</td>
    <td>0</td>
    <td>0</td>
    <td>...</td>
  </tr>
  <tr>
    <!--Russian words-->
    <td>0</td>
    <td>0</td>
    <td>0</td>
    <td>...</td>
    <!--Bulgarian words-->
    <td>0</td>
    <td>0</td>
    <td>0</td>
    <td>...</td>
    <!--...-->
    <td>...</td>
    <!--Polish words-->
    <td>5</td>
    <td>6</td>
    <td>3</td>
    <td>...</td>
  </tr>
  <tr>
    <!--Russian words-->
    <td>1</td>
    <td>0</td>
    <td>1</td>
    <td>...</td>
    <!--Bulgarian words-->
    <td>0</td>
    <td>1</td>
    <td>0</td>
    <td>...</td>
    <!--...-->
    <td>...</td>
    <!--Polish words-->
    <td>0</td>
    <td>1</td>
    <td>0</td>
    <td>...</td>
  </tr>
  <tr>
    <!--Russian words-->
    <td>0</td>
    <td>0</td>
    <td>0</td>
    <td>...</td>
    <!--Bulgarian words-->
    <td>0</td>
    <td>0</td>
    <td>0</td>
    <td>...</td>
    <!--...-->
    <td>...</td>
    <!--Polish words-->
    <td>0</td>
    <td>1</td>
    <td>0</td>
    <td>...</td>
  </tr>
</table>

## Which algorithms could be used to solve this problem?
I believe that a Decision Tree would be a good approach. Since the languages themselves are grouped in a tree-structure, and with a locally greedy approach, a Decision Tree could in theory model the historical linguistic development and development of different features well, thus modelling the structure of the problem itself. Of course, that similar languages share the same words is a big (and often false) assumption, together with the fact that the alphabet is not always dependent on the language group, means this would probably be very inaccurate, but an interesting experiment nonetheless.

For similar structural reasons I shall also use a Random Forest classifier, which is generally more accurate, however the resulting model is harder to for a human to interprate. 

For comparison, I will also train a Naive Bayes model on the data, however I am expecting it to be less accurate than the models described above. 

# Performance

## How can the performance be measured?
Splitting the corpus into training and test set allows the perfomance to be measured in a non-biased way. 

## And what performance is expected?
Given a text of sufficient length, I hope the accuracy will be over 90%. The main issue I could encounter, is that there a languages which have lots of words written then same way in common with each other, which could lead to misclassification. A way to combat this is to make sure both the list of most common words is long enough (i.e. that n is large enough) and that the test datapoints are long enough. 

## System Requirements
Python 3.9 (or similar)

## Author
Samuel Innes: dd257@stud.uni-heidelberg.de

## References
[^1]: Image source: http://strongwood.lv/slide-page/final_full_width/
[^2]: https://www.britannica.com/topic/Slavic-languages
[^3]: Image source: https://en.wikipedia.org/wiki/Slavic_languages
